"""Common python utilities for storage providers"""
from charmhelpers.core import hookenv
import os
import subprocess
import sys
from time import sleep
from stat import S_ISBLK, ST_MODE
import json

FSTAB_FILE = "/etc/fstab"
EC2_METADATA_URL = "http://169.254.169.254/latest/meta-data"
OS_METADATA_URL = "http://169.254.169.254/openstack/2012-08-10/meta_data.json"
PERSIST_DATA_FILE = "charm-data"


def log(message, level=None):
    """Quaint little log wrapper for juju logging"""
    hookenv.log(message, level)


# XXX We can remove _persist_data when juju bug:1279018 is fixed to provide
# relation-data from *-relation-departed hook runs
def _persist_data(key, value):
    """Save C{key} and C{value} pairs in a persistent file for reference"""
    data = {}
    filepath = "%s/%s" % (hookenv.charm_dir(), PERSIST_DATA_FILE)
    if os.path.exists(filepath):
        with open(filepath) as inputfile:
            data = json.load(inputfile)
    data[key] = value
    with open(filepath, "w") as outfile:
        outfile.write(unicode(json.dumps(data, ensure_ascii=False)))


def _get_persistent_data(key=None, remove_values=False):
    """Get a persistent data as specified by C{key} from the PERSIST_DATA_FILE

    If specified C{key} does not exist, or if no PERSIST_DATA_FILE exists,
    return C{None}. If C{remove_values} is C{True} remove the specified C{key}
    from the data file.
    """
    data = {}
    filepath = "%s/%s" % (hookenv.charm_dir(), PERSIST_DATA_FILE)
    if os.path.exists(filepath):
        with open(filepath) as inputfile:
            data = json.load(inputfile)
    if key:
        value = data.get(key, None)
        if remove_values:
            data.pop(key, None)
            if data:
                with open(filepath, "w") as outfile:
                    outfile.write(
                        unicode(json.dumps(data, ensure_ascii=False)))
            elif os.path.exists(filepath):
                os.remove(filepath)
    else:
        value = data
        if not data:
            value = None
        if remove_values:
            if os.path.exists(filepath):
                os.remove(filepath)
    return value


def get_provider():
    """Get the storage charm's provider type as set within it's config.yaml"""
    provider = hookenv.config("provider")
    if provider:
        return provider
    else:
        log(
            "Error: no provider defined in storage charm config",
            hookenv.ERROR)
        sys.exit(1)


def is_mounted(mountpoint):
    """Return True if the mountpoint is mounted"""
    if not mountpoint:
        log("INFO: No mountpoint given, assuming not mounted.")
        return False
    return os.path.ismount(mountpoint)


def unmount_volume():
    """Unmount the relation-requested C{mountpoint}"""
    success = False
    # Because relation-data is not guaranteed to exist on relation-departed
    # We need to obtain mountpoint from the persistent relation data we saved
    # during data-relation-changed
    mountpoint = _get_persistent_data("mountpoint")
    if not mountpoint:
        log("INFO: No mountpoint defined by relation no unmount performed.")
        return

    if not is_mounted(mountpoint):
        log("%s is not mounted, Done." % mountpoint)
        success = True
    else:
        for x in range(0, 20):
            if subprocess.call(["umount", mountpoint]) == 0:
                success = True
                log("Unmounted %s, Done" % mountpoint)
                break
            else:  # Then device is in use let's report the offending process
                message = "WARNING: umount %s failed. Retrying." % mountpoint
                try:
                    lsof = subprocess.check_output(["lsof", mountpoint])
                except subprocess.CalledProcessError, e:
                    log("WARNING: %s" % str(e), hookenv.WARNING)
                else:
                    process_name = lsof.split("\n")[1].split()[0]
                    message += " Device in use by (%s)" % process_name
                log(message, hookenv.WARNING)
            sleep(5)

    if not success:
        log(
            "ERROR: Unmount failed, leaving relation in errored state",
            hookenv.ERROR)
        sys.exit(1)


def _assert_block_device(device_path):
    """C{device_path} is a valid block device or exit in error"""
    isBlockDevice = False
    for x in range(0, 10):
        if not os.path.exists(device_path):
            log("WARNING: %s does not exist" % device_path)
        else:
            try:
                mode = os.stat(device_path)[ST_MODE]
                if S_ISBLK(mode):
                    log(
                        "DEBUG: block device found. Proceeding.",
                        hookenv.DEBUG)
                    isBlockDevice = True
                    break
            except OSError:
                pass
        sleep(5)
    if not isBlockDevice:
        log("ERROR: %s is not a block device." % device_path)
        sys.exit(1)


def _set_label(device_path, label):
    """Set label if necessary on C{device_path}"""
    if not os.path.exists(device_path):
        log("ERROR: Cannot set label of %s. Path does not exist" % device_path)
        sys.exit(1)

    current_label = subprocess.check_output(
        "e2label %s" % device_path, shell=True)
    if current_label:
        current_label = current_label.strip()
        if current_label == label:
            # Label already set, return
            return
        log("WARNING: %s had label=%s, overwriting with label=%s" %
            (device_path, current_label, label))
    subprocess.check_call(
        "e2label %s %s" % (device_path, label), shell=True)


def _read_partition_table(device_path):
    """Call blockdev to read a valid partition table and return exit code"""
    return subprocess.call(
        "blockdev --rereadpt %s" % device_path, shell=True)


def _assert_fstype(device_path, fstype):
    """Return C{True} if a filesystem is of type C{fstype}"""
    command = "file -s %s | egrep -q %s" % (device_path, fstype)
    return subprocess.call(command, shell=True) == 0


def _format_device(device_path):
    """
    Create an ext4 partition, if needed, for C{device_path} and fsck that
    partition.
    """
    result = _read_partition_table(device_path)
    if result == 1:
        log("INFO: %s is busy, no fsck performed. Assuming formatted." %
            device_path)
    elif result == 0:
        # Create an ext4 filesystem if NOT already present
        # use e.g. LABEl=vol-000012345
        if _assert_fstype(device_path, "ext4"):
            log("%s already formatted - skipping mkfs.ext4." % device_path)
        else:
            command = "mkfs.ext4 %s" % device_path
            log("NOTICE: Running: %s" % command)
            subprocess.check_call(command, shell=True)
        if subprocess.call("fsck -p %s" % device_path, shell=True) != 0:
            log("ERROR: fsck -p %s failed" % device_path)
            sys.exit(1)


def _get_from_relation(relation, key, default=None):
    relids = hookenv.relation_ids(relation)
    for relid in relids:
        for unit in hookenv.related_units(relid):
            relation = hookenv.relation_get(unit=unit, rid=relid)
            value = relation.get(key, default)
            if value is not None:
                return value
    return None


def mount_volume(device_path=None):
    """
    Mount the attached volume, nfs or block-storage-broker type, to the
    principal requested C{mountpoint}.

    If the C{mountpoint} required relation data does not exist, log the
    issue and exit. Otherwise attempt to initialize and mount the blockstorage
    or nfs device.
    """
    # mountpoint defined by our related principal in the "data" relation
    mountpoint = _get_from_relation("data", "mountpoint")

    if not mountpoint:
        log("No mountpoint defined by relation. Cannot mount volume yet.")
        return

    _persist_data("mountpoint", mountpoint)  # Saved for our departed hook
    if is_mounted(mountpoint):
        log("Volume (%s) already mounted" % mountpoint)
        # publish changes to data relation and exit
        for relid in hookenv.relation_ids("data"):
            hookenv.relation_set(
                relid, relation_settings={"mountpoint": mountpoint})
        return

    options = "default"
    provider = get_provider()
    if provider == "nfs":
        relids = hookenv.relation_ids("nfs")
        for relid in relids:
            for unit in hookenv.related_units(relid):
                relation = hookenv.relation_get(unit=unit, rid=relid)
                # XXX Only handle 1 related nfs unit. No nfs cluster support
                nfs_server = relation.get("private-address", None)
                nfs_path = relation.get("mountpoint", None)
                fstype = relation.get("fstype", None)
                options = relation.get("options", "default")
                break

        if len(relids) == 0 or None in [nfs_server, nfs_path, fstype]:
            log("Missing required relation values. "
                "nfs-relation-changed hook needs to run first.")
            return

        device_path = "%s:%s" % (nfs_server, nfs_path)
    elif provider == "block-storage-broker":
        if device_path is None:
            log("Waiting for persistent nova device provided by "
                "block-storage-broker.")
            sys.exit(0)

        _assert_block_device(device_path)
        if os.path.exists("%s1" % device_path):
            # Then we are supporting upgrade-charm path for deprecated
            # postgresql volume-map config parameter. Postgresql used to
            # partition the volume into a single partition at 100% of the
            # volume size, but the storage charm doesn't do that partitioning
            device_path = "%s1" % device_path
            log("Mounting %s as the device path as the root device is already "
                "partitioned." % device_path)
        _format_device(device_path)
        _set_label(device_path, mountpoint)
        fstype = subprocess.check_output(
            "blkid -o value -s TYPE %s" % device_path, shell=True).strip()
    else:
        log("ERROR: Unknown provider %s. Cannot mount volume." % hookenv.ERROR)
        sys.exit(1)

    if not os.path.exists(mountpoint):
        os.makedirs(mountpoint)

    options_string = "" if options == "default" else " -o %s" % options
    command = "mount -t %s%s %s %s" % (
        fstype, options_string, device_path, mountpoint)
    subprocess.check_call(command, shell=True)
    log("Mount (%s) successful" % mountpoint)
    with open(FSTAB_FILE, "a") as output_file:
        output_file.write(
            "%s %s %s %s 0 0" % (device_path, mountpoint, fstype, options))
    # publish changes to data relation and exit
    for relid in hookenv.relation_ids("data"):
        hookenv.relation_set(
            relid, relation_settings={"mountpoint": mountpoint})


def get_instance_id():
    """
    Curl metadata for openstack from C{OS_METADATA_URL} to obtain the C{uuid}.
    If that fails, we are likely on EC2 and curl the C{EC2_METADATA_URL} to get
    the EC2 C{instance-id}.
    """
    try:
        metadata = json.loads(
            subprocess.check_output("curl %s" % OS_METADATA_URL, shell=True))
        return metadata["uuid"]
    except ValueError:
        pass  # We are probably on EC2, try below
    try:
        metadata = subprocess.check_output(
            "curl %s/instance-id" % EC2_METADATA_URL, shell=True)
        return metadata.strip()
    except subprocess.CalledProcessError, e:
        log("Error: couldn't discover ec2 instance id. %s" % str(e),
            hookenv.ERROR)
        sys.exit(1)


def get_availability_zone():
    """
    Curl metadata for openstack from C{OS_METADATA_URL} to obtain the
    C{availability_zone}.
    If that fails, we are likely on EC2, so curl the
    C{EC2_METADATA_URL/placement/availability-zone} to get the EC2 availability
    zone.
    """
    try:
        metadata = json.loads(
            subprocess.check_output("curl %s" % OS_METADATA_URL, shell=True))
        return metadata["availability_zone"]
    except ValueError:
        pass  # We are probably on EC2, try below
    try:
        metadata = subprocess.check_output(
            "curl %s/placement/availability-zone" % EC2_METADATA_URL,
            shell=True)
        return metadata.strip()
    except subprocess.CalledProcessError, e:
        log("Error: couldn't discover ec2 availability zone. %s" % str(e),
            hookenv.ERROR)
        sys.exit(1)
