import common_util as util
import json
import mocker
import os
import subprocess
from testing import TestHookenv
from time import sleep


class TestCommonUtil(mocker.MockerTestCase):

    def setUp(self):
        self.maxDiff = None
        util.FSTAB_FILE = self.makeFile()
        util.hookenv = TestHookenv({"provider": "nfs"})
        self.addCleanup(setattr, os, "environ", os.environ)
        self.charm_dir = self.makeDir()
        os.environ = {"CHARM_DIR": self.charm_dir}

    def test_wb_persist_data_creates_persist_file_if_it_doesnt_exist(self):
        """
        L{_persist_data} will create a new PERSIST_DATA_FILE in C{CHARM_DIR}
        if the file does not already exist and store the provided C{key},
        C{value} pairs.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        self.assertFalse(os.path.exists(persist_path))
        util._persist_data("key", "value")
        with open(persist_path) as infile:
            data = json.load(infile)
            self.assertEqual(data["key"], "value")

    def test_wb_persist_data_appends_new_values_to_existing_persist_file(self):
        """
        L{_persist_data} will add new values to an existing PERSIST_DATA_FILE
        in C{CHARM_DIR} without removing existing C{key}, C{value} pairs.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        data = {"key1": "value1"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        util._persist_data("key2", "value2")
        with open(persist_path) as infile:
            data = json.load(infile)
            self.assertEqual(data["key1"], "value1")
            self.assertEqual(data["key2"], "value2")

    def test_wb_get_persistent_data_returns_none_when_no_persist_file(self):
        """
        L{_get_persistent_data} will return C{None} for any C{key} search if
        the PERSIST_DATA_FILE does not yet exist.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        self.assertFalse(os.path.exists(persist_path))
        self.assertIsNone(util._get_persistent_data("key2"))
        self.assertIsNone(util._get_persistent_data())

    def test_wb_get_persistent_data_returns_none_when_no_key_exists(self):
        """
        L{_get_persistent_data} will return C{None} for any C{key} search when
        the C{key} is not defined in the PERSIST_DATA_FILE.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        data = {"key1": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertIsNone(util._get_persistent_data("cant-find-this-key"))

    def test_wb_get_persistent_data_returns_value_when_key_exists(self):
        """
        L{_get_persistent_data} will return the corresponding value for the
        search C{key} provided when that C{key} is defined in the
        PERSIST_DATA_FILE.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        data = {"key1": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(util._get_persistent_data("key1"), "i-123")

    def test_wb_get_persistent_data_removes_key_when_remove_values_true(self):
        """
        L{_get_persistent_data} will return the corresponding value for the
        search C{key} provided when that C{key} is defined in the
        PERSIST_DATA_FILE. It will also remove that key from the
        PERSIST_DATA_FILE when C{remove_values} is C{True}.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        data = {"key1": "i-123", "key2": "i-456"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(
            util._get_persistent_data("key1", remove_values=True), "i-123")
        with open(persist_path) as infile:
            new_data = json.load(infile)
            self.assertEqual(new_data, {"key2": "i-456"})

    def test_wb_get_persistent_data_removes_file_when_remove_values_true(self):
        """
        L{_get_persistent_data} will return the corresponding value for the
        search C{key} provided when that C{key} is defined in the
        PERSIST_DATA_FILE. It will also remove the PERSIST_DATA_FILE when
        C{remove_values} is C{True} and the last remaining C{key} has been
        removed.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        data = {"key1": "i-123"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(
            util._get_persistent_data("key1", remove_values=True), "i-123")
        self.assertFalse(os.path.exists(persist_path))

    def test_wb_get_persistent_data_removes_file_when_no_key_specified(self):
        """
        L{_get_persistent_data} will return all existing values and remove the
        PERSIST_DATA_FILE when C{remove_values} is C{True} and no C{key} is
        specified.
        """
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        data = {"key1": "i-123", "key2": "i-456"}
        with open(persist_path, "w") as outfile:
            outfile.write(unicode(json.dumps(data, ensure_ascii=False)))
        self.assertEqual(
            util._get_persistent_data(key=None, remove_values=True),
            data)
        self.assertFalse(os.path.exists(persist_path))

    def test_get_provider(self):
        """L{get_provider} returns the C{provider} config setting if present"""
        self.assertEqual(util.get_provider(), "nfs")

    def test_get_provider_not_set(self):
        """
        L{get_provider} exits with an error if C{provider} configuration unset.
        """
        util.hookenv._config = {}  # No provider defined

        self.assertRaises(SystemExit, util.get_provider)
        message = "Error: no provider defined in storage charm config"
        self.assertTrue(util.hookenv.is_logged(message), "Error not logged")

    def test_get_instance_id_when_openstack(self):
        """
        L{get_instance_id} will attempt a curl for openstack metadata to
        extract the C{uuid} as the instance_id from openstack JSON metadata.
        """
        os_metadata = {
            "uuid": "i-123123"
        }
        curl = self.mocker.replace(subprocess.check_output)
        curl("curl %s" % util.OS_METADATA_URL, shell=True)
        self.mocker.result(json.dumps(os_metadata))
        self.mocker.replay()
        self.assertEqual(util.get_instance_id(), "i-123123")

    def test_get_instance_id_when_ec2(self):
        """
        L{get_instance_id} will first attempt a curl for openstack metadata.
        When that fails to be valid json, because we are on an EC2 instance,
        it will curl the C{EC2_METADATA_URL} to grab the C{instance-id}.
        """
        instance_id = "i-123123"
        curl = self.mocker.replace(subprocess.check_output)
        curl("curl %s" % util.OS_METADATA_URL, shell=True)
        self.mocker.throw(ValueError("invalid JSON"))
        curl("curl %s/instance-id" % util.EC2_METADATA_URL, shell=True)
        self.mocker.result(instance_id)
        self.mocker.replay()
        self.assertEqual(util.get_instance_id(), instance_id)

    def test_get_instance_id_fails_on_ec2(self):
        """
        When L{get_instance_id} fails to curl C{EC2_METADATA_URL} it will log
        the error and exit in error.
        """
        ec2_command = "curl %s/instance-id" % util.EC2_METADATA_URL

        curl = self.mocker.replace(subprocess.check_output)
        curl("curl %s" % util.OS_METADATA_URL, shell=True)
        self.mocker.throw(ValueError("invalid JSON"))
        curl(ec2_command, shell=True)
        self.mocker.throw(
            subprocess.CalledProcessError(1, ec2_command))
        self.mocker.replay()
        result = self.assertRaises(SystemExit, util.get_instance_id)
        self.assertEqual(result.code, 1)
        message = (
            "Error: couldn't discover ec2 instance id. Command '%s' returned "
            "non-zero exit status 1" % ec2_command)
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_get_availability_zone_when_openstack(self):
        """
        L{get_availability_zone} will attempt a curl for openstack metadata to
        extract the C{uuid} as the instance_id from openstack JSON metadata.
        """
        os_metadata = {
            "availability_zone": "nova"
        }
        curl = self.mocker.replace(subprocess.check_output)
        curl("curl %s" % util.OS_METADATA_URL, shell=True)
        self.mocker.result(json.dumps(os_metadata))
        self.mocker.replay()
        self.assertEqual(util.get_availability_zone(), "nova")

    def test_get_availability_zone_when_ec2(self):
        """
        L{get_availability_zone} will first attempt a curl for openstack
        metadata. When that fails to be valid JSON because we are on an EC2
        instance, it will curl the C{EC2_METADATA_URL} to grab the
        C{availability-zone}.
        """
        availability_zone = "us-east-1b"
        curl = self.mocker.replace(subprocess.check_output)
        curl("curl %s" % util.OS_METADATA_URL, shell=True)
        self.mocker.throw(ValueError("invalid JSON"))
        curl("curl %s/placement/availability-zone" %
             util.EC2_METADATA_URL, shell=True)
        self.mocker.result(availability_zone)
        self.mocker.replay()
        self.assertEqual(util.get_availability_zone(), availability_zone)

    def test_get_availability_zone_fails_on_ec2(self):
        """
        When L{get_availability_zone} fails to curl C{EC2_METADATA_URL} it will
        log the error and exit in error.
        """
        ec2_command = ("curl %s/placement/availability-zone" %
                       util.EC2_METADATA_URL)

        curl = self.mocker.replace(subprocess.check_output)
        curl("curl %s" % util.OS_METADATA_URL, shell=True)
        self.mocker.throw(ValueError("invalid JSON"))
        curl(ec2_command, shell=True)
        self.mocker.throw(
            subprocess.CalledProcessError(1, ec2_command))
        self.mocker.replay()
        result = self.assertRaises(SystemExit, util.get_availability_zone)
        self.assertEqual(result.code, 1)
        message = (
            "Error: couldn't discover ec2 availability zone. Command '%s' "
            "returned non-zero exit status 1" % ec2_command)
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_is_mounted_when_mountpoint_is_mounted(self):
        """
        L{is_mounted} returns C{True} when the mountpath is mounted.
        """
        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(True)
        self.mocker.replay()
        self.assertTrue(util.is_mounted("/mnt/this"))

    def test_is_mounted_when_mountpoint_is_not_mounted(self):
        """
        L{is_mounted} returns C{False} when the mountpath is not mounted.
        """
        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(False)
        self.mocker.replay()

        self.assertFalse(util.is_mounted("/mnt/this"))

    def test_is_mounted_when_mountpoint_not_defined(self):
        """
        L{is_mounted} returns C{False} when the given mountpoint is undefined.
        """
        self.assertIsNone(util.hookenv.relation_get("mountpoint"))
        self.assertFalse(util.is_mounted(None))
        message = "No mountpoint given, assuming not mounted."
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_unmount_volume_when_mountpoint_none(self):
        """
        L{unmount_volume} will log a message and do nothing if mountpoint is
        not defined.
        """
        self.assertIsNone(util._get_persistent_data("mountpoint"))
        util.unmount_volume()

        message = "No mountpoint defined by relation no unmount performed."
        self.assertTrue(util.hookenv.is_logged(message), "Message not logged")

    def test_unmount_volume_when_defined_mountpoint_not_mounted(self):
        """
        L{unmount_volume} will log a message and do nothing if the defined
        mountpoint is not mounted.
        """
        util._persist_data("mountpoint", "/mnt/this")

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(False)
        self.mocker.replay()

        util.unmount_volume()

        message = "/mnt/this is not mounted, Done."
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_unmount_volume_multiple_failure(self):
        """
        L{unmount_volume} will attempt to unmount a mounted filesystem 20 times
        to get a success with a sleep(5) in between attempts. If all attempts
        fail, L{umount_volume} will log the failure and exit 1.
        """
        util._persist_data("mountpoint", "/mnt/this")

        lsof_output = (
            "COMMAND     PID   USER   FD   TYPE DEVICE SIZE/OFF  NODE NAME\n"
            "postgresql 30544 csmith   10r   CHR    1,9    0t0  2056 "
            "/mnt/this")

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(True)
        umount = self.mocker.replace(subprocess.call)
        umount(["umount", "/mnt/this"])
        self.mocker.result(2)  # Failure from the shell cmd
        self.mocker.count(20)
        lsof = self.mocker.replace(subprocess.check_output)
        lsof(["lsof", "/mnt/this"])
        self.mocker.result(lsof_output)
        self.mocker.count(20)
        sleep_mock = self.mocker.replace(sleep)
        sleep_mock(5)
        self.mocker.count(20)
        self.mocker.replay()

        self.assertRaises(SystemExit, util.unmount_volume)
        message = "ERROR: Unmount failed, leaving relation in errored state"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_unmount_lsof_command_failure(self):
        """Log a warning if lsof fails during L{unmount_volume} retries"""
        util._persist_data("mountpoint", "/mnt/this")

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(True)
        umount = self.mocker.replace(subprocess.call)
        umount(["umount", "/mnt/this"])
        self.mocker.result(2)  # Failure from the umount shell cmd
        self.mocker.count(20)
        lsof = self.mocker.replace(subprocess.check_output)
        lsof(["lsof", "/mnt/this"])
        self.mocker.count(20)
        self.mocker.throw(
            subprocess.CalledProcessError(1, "lsof /mnt/this"))  # lsof fail
        sleep_mock = self.mocker.replace(sleep)
        sleep_mock(5)
        self.mocker.count(20)
        self.mocker.replay()

        self.assertRaises(SystemExit, util.unmount_volume)
        messages = [
            "ERROR: Unmount failed, leaving relation in errored state",
            "WARNING: umount /mnt/this failed. Retrying.",
            "WARNING: Command 'lsof /mnt/this' returned non-zero exit status "
            "1"]
        for message in messages:
            self.assertTrue(
                util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_unmount_volume_success(self):
        """
        L{unmount_volume} will unmount a mounted filesystem calling the shell
        command umount and will log the success.
        """
        util._persist_data("mountpoint", "/mnt/this")

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(True)
        unmount_cmd = self.mocker.replace(subprocess.call)
        unmount_cmd(["umount", "/mnt/this"])
        self.mocker.result(0)  # Success from the shell cmd
        self.mocker.replay()

        util.unmount_volume()
        message = "Unmounted /mnt/this, Done"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_assert_block_device_failure_path_does_not_exist(self):
        """
        L{_assert_block_device} will exit in error if the C{device_path}
        provided does not exist.
        """
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdx")
        self.mocker.result(False)
        self.mocker.count(10)
        sleep_mock = self.mocker.replace(sleep)
        sleep_mock(5)
        self.mocker.count(10)
        self.mocker.replay()

        self.assertRaises(SystemExit, util._assert_block_device, "/dev/vdx")
        messages = ["WARNING: /dev/vdx does not exist",
                    "ERROR: /dev/vdx is not a block device."]
        for message in messages:
            self.assertTrue(
                util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_assert_block_device_failure_not_block_device(self):
        """
        L{_assert_block_device} will try 10 times between sleeps to validate
        that C{device_path} is a valid block device before logging an error
        and exiting.
        """
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdx")
        self.mocker.result(True)  # Mock that file exists
        self.mocker.count(10)
        stat = self.mocker.replace(os.stat)
        stat("/dev/vdx")
        self.mocker.count(10)
        self.mocker.throw(OSError)  # When device file doesn't exist
        sleep_mock = self.mocker.replace(sleep)
        sleep_mock(5)
        self.mocker.count(10)
        self.mocker.replay()

        self.assertRaises(SystemExit, util._assert_block_device, "/dev/vdx")
        message = "ERROR: /dev/vdx is not a block device."
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_set_label_exits_when_path_does_not_exist(self):
        """
        L{_set_label} will log and exit when C{device_path} does not exist
        """
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/not-there")
        self.mocker.result(False)
        self.mocker.replay()
        self.assertRaises(
            SystemExit, util._set_label, "/dev/not-there", "/mnt/this")
        message = (
            "ERROR: Cannot set label of /dev/not-there. Path does not exist")
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_set_label_does_nothing_if_label_already_set_correctly(self):
        """
        L{_set_label} will use e2label command to check if the label is
        set. If current label is equal to requested label. L{_set_label} will
        do nothing.
        """
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdz")
        self.mocker.result(True)
        e2label_check = self.mocker.replace(subprocess.check_output)
        e2label_check("e2label /dev/vdz", shell=True)
        self.mocker.result("/mnt/this\n")
        self.mocker.replay()
        util._set_label("/dev/vdz", "/mnt/this")

    def test_wb_set_label_alters_existing_label_when_set_different(self):
        """
        L{_set_label} will use e2label command to check if the label is
        set. If current label is different from requested label. L{_set_label}
        will reset the filesystem label and log a message.
        """
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdz")
        self.mocker.result(True)
        e2label_check = self.mocker.replace(subprocess.check_output)
        e2label_check("e2label /dev/vdz", shell=True)
        self.mocker.result("other-label\n")
        e2label_check = self.mocker.replace(subprocess.check_call)
        e2label_check("e2label /dev/vdz /mnt/this", shell=True)
        self.mocker.replay()
        util._set_label("/dev/vdz", "/mnt/this")
        message = (
            "WARNING: /dev/vdz had label=other-label, overwriting with "
            "label=/mnt/this")
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_set_label_alters_existing_label_when_unset(self):
        """
        L{_set_label} will use e2label command to check if the label is
        set. If label is unset, L{_set_label} to reset the filesystem label and
        log a message.
        """
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdz")
        self.mocker.result(True)
        e2label_check = self.mocker.replace(subprocess.check_output)
        e2label_check("e2label /dev/vdz", shell=True)
        self.mocker.result("")
        e2label_check = self.mocker.replace(subprocess.check_call)
        e2label_check("e2label /dev/vdz /mnt/this", shell=True)
        self.mocker.replay()
        util._set_label("/dev/vdz", "/mnt/this")

    def test_wb_format_device_ignores_fsck_when_busy(self):
        """
        When blockdev reports the C{device_path} is busy,
        L{_format_device} skips fsck and returns the partition path.
        """
        partition_table_mock = self.mocker.replace(util._read_partition_table)
        partition_table_mock("/dev/vdz")
        self.mocker.result(1)
        self.mocker.replay()

        util._format_device("/dev/vdz")
        message = (
            "INFO: /dev/vdz is busy, no fsck performed. Assuming formatted.")
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_format_device_runs_fsck_if_device_not_busy(self):
        """
        When L{_read_partition_table} reports the device partition table is
        readable, and L{_assert_fstype} returns C{True}, mkfs.ext4 is not run
        but fsck is run.
        """
        partition_table_mock = self.mocker.replace(util._read_partition_table)
        partition_table_mock("/dev/vdz")
        self.mocker.result(0)  # Report that partition table is readable

        # Check formatted filesystem type
        ext4_check = self.mocker.replace(util._assert_fstype)
        ext4_check("/dev/vdz", "ext4")
        self.mocker.result(True)

        # Assert fsck is called
        fsck = self.mocker.replace(subprocess.call)
        fsck("fsck -p /dev/vdz", shell=True)
        self.mocker.result(0)   # successful fsck command exit
        self.mocker.replay()

        util._format_device("/dev/vdz")
        message = "/dev/vdz already formatted - skipping mkfs.ext4."
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_format_device_non_ext4_is_formatted(self):
        """
        When L{_read_partition_table} reports the device partition table is
        readable, and L{_assert_fstype} returns C{Flase}, mkfs.ext4 and fsck
        are run.
        """
        partition_table_mock = self.mocker.replace(util._read_partition_table)
        partition_table_mock("/dev/vdz")
        self.mocker.result(0)  # Report that partition table is readable

        # Check formatted filesystem type
        ext4_check = self.mocker.replace(util._assert_fstype)
        ext4_check("/dev/vdz", "ext4")
        self.mocker.result(False)
        fsck = self.mocker.replace(subprocess.check_call)
        fsck("mkfs.ext4 /dev/vdz", shell=True)
        fsck = self.mocker.replace(subprocess.call)
        fsck("fsck -p /dev/vdz", shell=True)
        self.mocker.result(0)   # Sucessful command exit 0
        self.mocker.replay()

        util._format_device("/dev/vdz")
        message = "NOTICE: Running: mkfs.ext4 /dev/vdz"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_wb_format_device_responds_to_fsck_error(self):
        """
        When fsck returns an error, L{_format_device} exits in
        error and logs the error.
        """
        partition_table_mock = self.mocker.replace(util._read_partition_table)
        partition_table_mock("/dev/vdz")
        self.mocker.result(0)  # Report that partition table is readable

        # Check formatted filesystem type
        ext4_check = self.mocker.replace(util._assert_fstype)
        ext4_check("/dev/vdz", "ext4")
        self.mocker.result(False)
        fsck = self.mocker.replace(subprocess.check_call)
        fsck("mkfs.ext4 /dev/vdz", shell=True)
        fsck = self.mocker.replace(subprocess.call)
        fsck("fsck -p /dev/vdz", shell=True)
        self.mocker.result(1)
        self.mocker.replay()

        result = self.assertRaises(
            SystemExit, util._format_device, "/dev/vdz")
        self.assertEqual(result.code, 1)

        messages = ["NOTICE: Running: mkfs.ext4 /dev/vdz",
                    "ERROR: fsck -p /dev/vdz failed"]
        for message in messages:
            self.assertTrue(
                util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_mount_volume_mountpoint_not_in_relation(self):
        """
        L{mount_volume} will log a message when called while the C{mountpoint}
        has not yet defined by the relation principal yet.
        """
        self.assertIsNone(util.hookenv.relation_get("mountpoint"))
        util.mount_volume()
        message = "No mountpoint defined by relation. Cannot mount volume yet."
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_mount_volume_mountpoint_persists_mountpoint_from_relation(self):
        """
        L{mount_volume} will read incoming principal relation data for the
        C{mountpoint}. When the requested C{mountpoint} is determined as
        mounted,  it will then persist the mountpoint data, log a message and
        set juju relation C{mountpoint} in response.
        """
        util.hookenv._incoming_relation_data = (
            ("data:1", "mountpoint", "/mnt/this"),)

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(True)
        self.mocker.replay()

        util.mount_volume()
        self.assertIn(("mountpoint", "/mnt/this"),
                      util.hookenv._outgoing_relation_data)
        message = "Volume (/mnt/this) already mounted"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        with open(persist_path) as infile:
            data = json.load(infile)
            self.assertEqual(data["mountpoint"], "/mnt/this")

    def test_mount_volume_mountpoint_from_relation_data_already_mounted(self):
        """
        L{mount_volume} will log a message and return when the C{mountpoint} is
        already defined in relation data and the requested C{mountpoint} is
        already mounted.
        """
        util.hookenv._incoming_relation_data = (
            ("data:1", "mountpoint", "/mnt/this"),)
        self.assertEqual(
            util.hookenv.relation_get("mountpoint", rid="data:1"), "/mnt/this")

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(True)
        self.mocker.replay()

        util.mount_volume()
        message = "Volume (/mnt/this) already mounted"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_mount_volume_nfs_provider_no_nfs_hook_relation_data(self):
        """
        L{mount_volume} will exit in error when set as C{nfs} provider and the
        nfs relation data does not contain any of the required
        values: C{mount_path}, C{mount_server}, C{mount_options} or
        C{mount_type}.
        """
        data = (
            ("data:1", "mountpoint", "/mnt/this"),
            ("nfs:1", "private-address", "me.com"),
            ("nfs:1", "mountpoint", "/nfs/server/path"),
            ("nfs:1", "fstype", "nfs"),)
        util.hookenv._incoming_relation_data = data

        # Setup persisted data missing values from nfs-relation-changed
        for skip in range(1, 3):
            partial_data = ()
            for index, item in enumerate(data):
                if index != skip:
                    partial_data = partial_data + (item,)
            util.hookenv._incoming_relation_data = partial_data
            self.assertEqual(
                util.hookenv.relation_get("mountpoint", rid="data:1"),
                "/mnt/this")
            util.mount_volume()
            message = (
                "Missing required relation values. "
                "nfs-relation-changed hook needs to run first.")
            self.assertTrue(
                util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_mount_volume_nfs_provider_success(self):
        """
        L{mount_volume} will read incoming principal relation data for the
        C{mountpoint}.  When storage config is set to the C{nfs} provider
        L{mount_volume} as a result of the hook C{nfs-relation-changed}. The
        values required nfs relation data are C{mount_path}, C{mount_server},
        C{mount_type} and C{mount_options}. The success of the mount will log a
        message and the storage charm will call juju relation-set C{mountpoint}
        to report the successful initialization to the principal.
        """
        util.hookenv._incoming_relation_data = (
            ("data:1", "mountpoint", "/mnt/this"),
            ("nfs:1", "private-address", "me.com"),
            ("nfs:1", "options", "someopts"),
            ("nfs:1", "mountpoint", "/nfs/server/path"),
            ("nfs:1", "fstype", "nfs"))

        # Mock not mounted
        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(False)

        # mount_volume calls makedirs when path !exists
        exists = self.mocker.replace(os.path.exists)
        exists("/mnt/this")
        self.mocker.result(False)
        makedirs = self.mocker.replace(os.makedirs)
        makedirs("/mnt/this")

        # Called proper mount command based on nfs data
        command = "mount -t nfs -o someopts me.com:/nfs/server/path /mnt/this"
        mount_cmd = self.mocker.replace(subprocess.check_call)
        mount_cmd(command, shell=True)
        self.mocker.replay()

        self.assertEqual(util.get_provider(), "nfs")
        util.mount_volume()
        message = "Mount (/mnt/this) successful"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

        # Wrote proper fstab mount info to mount on reboot
        fstab_content = "me.com:/nfs/server/path /mnt/this nfs someopts 0 0"
        with open(util.FSTAB_FILE) as input_file:
            self.assertEqual(input_file.read(), fstab_content)

        # Reported success to principal
        self.assertEqual(
            util.hookenv._outgoing_relation_data,
            (("mountpoint", "/mnt/this"),))
        # Perserved mountpoint for future departed hook runs
        persist_path = "%s/%s" % (self.charm_dir, util.PERSIST_DATA_FILE)
        with open(persist_path) as infile:
            data = json.load(infile)
            self.assertEqual(data["mountpoint"], "/mnt/this")

    def test_mount_volume_bsb_provider_no_device_path_provided(self):
        """
        L{mount_volume} will exit early if called before C{device_path}
        is specified by the block-storage-broker charm.
        """
        util.hookenv._config = (("provider", "block-storage-broker"),)
        util.hookenv._incoming_relation_data = (
            ("data:1", "mountpoint", "/mnt/this"),)

        self.assertEqual(util.get_provider(), "block-storage-broker")
        result = self.assertRaises(SystemExit, util.mount_volume)
        self.assertEqual(result.code, 0)
        message = (
            "Waiting for persistent nova device provided by "
            "block-storage-broker.")
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

    def test_mount_volume_bsb_provider_success(self):
        """
        L{mount_volume} reads principal relation data for the C{mountpoint}.
        When storage config is set to the C{nova} provider, L{mount_volume}
        will read the optional C{device_path} parameter provided by the nova
        provider's C{data-relation-changed} hook.
        The success of the mount will log a message and the storage charm will
        call relation-set C{mountpoint} to report the successful mount to the
        principal.
        """
        util.hookenv._incoming_relation_data = (
            ("data:1", "mountpoint", "/mnt/this"),)

        util.hookenv._config = (("provider", "block-storage-broker"),)

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(False)  # Mock not mounted

        assert_device = self.mocker.replace(util._assert_block_device)
        assert_device("/dev/vdx")
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdx1")
        self.mocker.result(False)
        create_partition = self.mocker.replace(
            util._format_device)
        create_partition("/dev/vdx")

        _set_label = self.mocker.replace(util._set_label)
        _set_label("/dev/vdx", "/mnt/this")

        # Ensure blkid is called to discover fstype of nova device
        get_fstype = self.mocker.replace(subprocess.check_output)
        get_fstype("blkid -o value -s TYPE /dev/vdx", shell=True)
        self.mocker.result("ext4\n")

        exists("/mnt/this")
        self.mocker.result(True)

        mount = self.mocker.replace(subprocess.check_call)
        mount("mount -t ext4 /dev/vdx /mnt/this", shell=True)
        self.mocker.replay()

        self.assertEqual(util.get_provider(), "block-storage-broker")
        util.mount_volume("/dev/vdx")
        message = "Mount (/mnt/this) successful"
        self.assertTrue(
            util.hookenv.is_logged(message), "Not logged- %s" % message)

        # Wrote proper fstab mount info to mount on reboot
        fstab_content = "/dev/vdx /mnt/this ext4 default 0 0"
        with open(util.FSTAB_FILE) as input_file:
            self.assertEqual(input_file.read(), fstab_content)

        # Reported success to principal
        self.assertEqual(
            util.hookenv._outgoing_relation_data,
            (("mountpoint", "/mnt/this"),))

    def test_mount_volume_bsb_provider_success_postgresql_charm_upgrade(self):
        """
        Old versions of the postgresql charm would have created a single
        partition with max size from an external attached volume if the user
        had used postgresql charm's deprecated volume-map parameter. When the
        storage charm sees a single partition on the attached device, it will
        attempt to mount that partition instead of the root device.
        The success of the mount will log a message and the storage charm will
        call relation-set C{mountpoint} to report the successful mount to the
        principal.
        """
        util.hookenv._incoming_relation_data = (
            ("data:1", "mountpoint", "/mnt/this"),)

        util.hookenv._config = (("provider", "block-storage-broker"),)

        ismount = self.mocker.replace(os.path.ismount)
        ismount("/mnt/this")
        self.mocker.result(False)  # Mock not mounted

        assert_device = self.mocker.replace(util._assert_block_device)
        assert_device("/dev/vdx")
        exists = self.mocker.replace(os.path.exists)
        exists("/dev/vdx1")
        self.mocker.result(True)
        create_partition = self.mocker.replace(
            util._format_device)
        create_partition("/dev/vdx1")

        _set_label = self.mocker.replace(util._set_label)
        _set_label("/dev/vdx1", "/mnt/this")

        # Ensure blkid is called to discover fstype of nova device
        get_fstype = self.mocker.replace(subprocess.check_output)
        get_fstype("blkid -o value -s TYPE /dev/vdx1", shell=True)
        self.mocker.result("ext4\n")

        exists("/mnt/this")
        self.mocker.result(True)

        mount = self.mocker.replace(subprocess.check_call)
        mount("mount -t ext4 /dev/vdx1 /mnt/this", shell=True)
        self.mocker.replay()

        self.assertEqual(util.get_provider(), "block-storage-broker")
        util.mount_volume("/dev/vdx")
        messages = ["Mount (/mnt/this) successful",
                    "Mounting /dev/vdx1 as the device path as the root device "
                    "is already partitioned."]
        for message in messages:
            self.assertTrue(
                util.hookenv.is_logged(message), "Not logged- %s" % message)

        # Wrote proper fstab mount info to mount on reboot
        fstab_content = "/dev/vdx1 /mnt/this ext4 default 0 0"
        with open(util.FSTAB_FILE) as input_file:
            self.assertEqual(input_file.read(), fstab_content)

        # Reported success to principal
        self.assertEqual(
            util.hookenv._outgoing_relation_data,
            (("mountpoint", "/mnt/this"),))
